/*-
 * SYNOPSIS:
 * int vtop(paddr_t v, paddr_t *p)
 *   Stores the physical address corresponding to virtual address 'v'
 *   in *p.
 *
 * RETURNS:
 *   EINVAL - virtual address is invalid.
 *   ENXIO  - virtual address has no mapping.
 *   EOK    - sucessfull.
 *
 * BUGS:
 *   is dependent upon QNX page layout.
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <i86.h>

typedef union pte pte_t;

union pte {
	ulong_t           pg_ival;
	struct {
		ulong_t       pg_present:1,
		                        :11,
                      pg_pfn    :20;
	};
};
#define PTAB_SIZE       (__PAGESIZE*__PAGESIZE/sizeof(pte_t))
#define PDIRP(x) ((pte_t __far *)MK_FP(0x20, 0xfffff000) + (x)/PTAB_SIZE)
#define PTEP(x)  ((pte_t __far *)MK_FP(0x20, 0xffc00000) + (x)/__PAGESIZE)

static int
get_pte(paddr_t va, pte_t *p)
{
	if (PDIRP(va)->pg_present == 0)
		return -1;
	*p = *PTEP(va);
	return 0;
}

int
vtop(paddr_t v, paddr_t *p)
{
	pte_t x;
	if (get_pte(v, &x) == -1)
		return EINVAL;
	if (!x.pg_present)
		return ENXIO;
	*p = x.pg_pfn*__PAGESIZE | (v&(__PAGESIZE-1));
	return EOK;
}


#ifdef TESTING

int         x;
#include <sys/kernel.h>
#include <sys/seginfo.h>
#include <sys/inline.h>

int
slow_vtop(paddr_t va, paddr_t *pa)
{
	int      x;
	int      sel;
	struct _seginfo y;
	sel = qnx_segment_overlay_flags(va & ~(__PAGESIZE-1), __PAGESIZE,
									_PMF_VOVERLAY);
	if (sel == -1)
		return errno;
	if (qnx_segment_info(PROC_PID, 0, sel, &y) != -1) {
		x = EOK;
		*pa = y.addr + (va & (__PAGESIZE-1));
	} else {
		x = errno;
	}
	qnx_segment_free(sel);
	return x;
}

main()
{
	struct _seginfo y;
	paddr_t px;
	/*-
	 * stdio may malloc, which may move our ds.  Make sure it
	 * moves first.
	 */
	printf("%s", " \b");
	qnx_segment_info(PROC_PID, 0, _ds(), &y);
	if ((errno=vtop(y.addr+(paddr_t)&x, &px)) != EOK) {
		perror("vtop:");
		exit(1);
	}
	printf("&x = %x, vaddr = %x, paddr %x\n", &x, y.addr+(paddr_t)&x, px);
	if ((errno=slow_vtop(y.addr+(paddr_t)&x, &px)) != EOK) {
		perror("slow_vtop:");
		exit(1);
	}
	printf("&x = %x, vaddr = %x, paddr %x\n", &x, y.addr+(paddr_t)&x, px);
}

#endif
